#include "cmd_join.h"

/* cmd_join - deals with join commands */    
int cmd_join(struct clientinfo *sender,struct parsed_cmd *cmd)
{
	/* Max size allowed by irc */
	char buffer[BUF_SIZE];
	struct channel *new_chan,*chan_info;
	int exists=0;
	buffer[0]='\0';
	
	/* Check for params */
	if(cmd->params[0]!=NULL) {
		/* Check channel name is valid */
		if(!check_channel(cmd->params[0])) {
			ircd_strncat(buffer,cmd->params[0],BUF_SIZE);
			ircd_strncat(buffer," :Bad Channel Mask",BUF_SIZE);
			send_numeric(sender->fd,sender->nick,476,buffer);
			cmd_free(cmd);
			return EXIT_FAILURE;	
		}
		
		/* Check for JOIN 0, which is a PART */
		if(lc_strncmp(cmd->params[0],"0",-1)) {
			/* return cmd_part; */
		}
		/* attempt to locate channel */
		chan_info=locate_channel(cmd->params[0]);
		/* if exists=1, then ch_channel points to the channel we want to join */
		/* otherwise we have to create the channel */
		/* need some error checking in here */
		if(chan_info!=NULL) {
			if(is_member(sender,chan_info)) {
				add_client_to_chan(sender,chan_info,0);
				inform_room_join(sender,chan_info);
				confirm_join(sender,chan_info);
				/* return */
				return cmd_free(cmd);
			} else {
				/* Client is already in that room */
				return cmd_free(cmd);
			}
		} else {
			/* Create the channel */
			new_chan=create_channel(cmd->params[0]);
			if(new_chan!=NULL) {
				add_client_to_chan(sender,new_chan,1);
				inform_room_join(sender,new_chan);
				confirm_join(sender,new_chan);
				return cmd_free(cmd);
			} else {
				/* Failed to create channel */
				/* Inform user */
				return cmd_free(cmd);
			}
		}
				
	} else {
		/* Inform client */
		send_numeric(sender->fd,sender->nick,461,"JOIN :Not enough parameters");
		return cmd_free(cmd);
	}
}

/* Updates linked lists */
/* needs to check malloc retvals */
static int add_client_to_chan(struct clientinfo *joinme,struct channel *channel,int op)
{
	struct user_wrap *curr_client;
	struct chan_wrap *curr_chan;
	struct user_wrap *prev_client;
	struct chan_wrap *prev_chan;
	struct user_wrap *new_client;
	struct chan_wrap *new_chan;
	
	/* Locate free space on the end of channel user list */
	curr_client=channel->users_start;	
	while(curr_client!=NULL) {
		prev_client=curr_client;
		curr_client=curr_client->next;
	}
	
	/* Allocate space for client wrapper */
	new_client=malloc(sizeof(struct user_wrap));
	if(new_client==NULL) 
		return EXIT_FAILURE;	
	
	/* Actually Add client */	
	if(channel->users_start==curr_client) {
		channel->users_start=new_client;
	} else {
		prev_client->next=new_client;
	}
	new_client->user_info=joinme;
	new_client->next=NULL;
	
	/* Set channel modes for user */
	new_client->chan_mode = CHANUSR_VOICE;
	/* If user is first into group, give them ops */
	if(op==1) 
		new_client->chan_mode |= CHANUSR_OP;

	/* Add channel on end of client list */
	curr_chan=joinme->chan_start;
	new_chan=malloc(sizeof(struct chan_wrap));
	while(curr_chan!=NULL) {
		prev_chan=curr_chan;
		curr_chan=curr_chan->next;
	}
	
	if(curr_chan==joinme->chan_start) {
		joinme->chan_start=new_chan;
	} else {
		prev_chan->next=new_chan;
	}
	
	new_chan->chan_info=channel;
	new_chan->next=NULL;	
}

/* informs members of a room about new member */
/* purposely includes person who just joined */
static int inform_room_join(struct clientinfo *joiner,struct channel *chan)
{
	struct user_wrap *curr_client;
	char buffer[BUF_SIZE]="JOIN";
	
	ircd_sprintf(buffer,BUF_SIZE,"JOIN %s",chan->name);
	send_to_channel(joiner,chan,buffer);
	
}

/* Creates a channel, returns a pointer to it */
/* takes name and position as args */
static struct channel *create_channel(char * name)
{
	struct channel *new_chan;

	new_chan=malloc(sizeof(struct channel));

	if(new_chan!=NULL) {
		/* Alloc'ed memory successfully */
		strncpy(new_chan->name,name,CHANNEL_MAX_LENGTH);
		new_chan->topic_settime=0;
		new_chan->topic_set[0]='\0';
		new_chan->topic[0]='\0';
		new_chan->mode = CMODE_TOPIC | CMODE_MEMONLYSEND;
		new_chan->bans_start = NULL;
		add_channel(new_chan);
		return new_chan;
	} else { 
		return NULL;
	}
}

/* user has joined room, send them topic and names list */
static int confirm_join(struct clientinfo *client,struct channel *chan) 
{
	char buffer[BUF_SIZE];
	struct user_wrap * curr_user;
	int start;
	buffer[0]='\0';
	
	/* Send topic */
	if(chan->topic[0]!='\0') {
		/* Send topic */
		ircd_sprintf(buffer,BUF_SIZE,"%s :%s",chan->name,chan->topic);
		send_numeric(client->fd,client->nick,332,buffer);
		/* Who set it */
		ircd_sprintf(buffer,BUF_SIZE,"%s %s %ld",chan->topic_set,chan->topic_settime);
		send_numeric(client->fd,client->nick,333,buffer);
	} else {
		/* No Topic is set */
		ircd_sprintf(buffer,BUF_SIZE,"%s :No topic is set",chan->name);
		send_numeric(client->fd,client->nick,331,buffer);
	}

	/* Names list */
	sprintf(buffer,"= %s :",chan->name);
	curr_user=chan->users_start;
	start=1;
	while(curr_user!=NULL) {
		if(start==1) {
			sprintf(buffer,"%s%s%s",buffer,
						(curr_user->chan_mode & CHANUSR_OP)?"@":"",
						curr_user->user_info->nick);
			start=0;
		} else {
			sprintf(buffer,"%s%s%s",buffer,
						(curr_user->chan_mode & CHANUSR_OP)?" @":" ",
						curr_user->user_info->nick);
		}
		curr_user=curr_user->next;
	}
	send_numeric(client->fd,client->nick,353,buffer);
	sprintf(buffer,"%s :End of /NAMES list.",chan->name);
	send_numeric(client->fd,client->nick,366,buffer);
}

/* check channel format */
int check_channel(char *chan_name)
{
	int i;
	int chan_len;
	
	
	/* Check first param is correct */
	switch(chan_name[0]) {

		case '#' : case '+' : case '&' : case '!' :
			/* Valid start characters */
			break;
		
		default:
			return 0;
			break;
	}
	
	chan_len=strlen(chan_name);
	
	/* Check channel length */
	if(chan_len >  CHANNEL_MAX_LENGTH) 
		return 0;

	/* Search for invalid characters */
	for(i=0;i < chan_len;i++) {
		switch(chan_name[i]) {

			case ' ' : case	7 : case ',' :
				return 0;
				break;

			default:
				break;
		}
	}

	return 1;
}

/* check a user isn't already in a channel */
static int is_member(struct clientinfo *client,struct channel *chan)
{
	struct user_wrap *curr_user;

	curr_user=chan->users_start;

	while(curr_user!=NULL) {
		if(curr_user->user_info==client) {
			/* They are a member */
			return 0;
		} else {
			curr_user=curr_user->next;
		}
	}
	/* False */
	return 1;
}
