#include "ircd.h"

long int clients=0;
long int ops=0;
int channels=0;
int services=0;
int signal_caught=0;

/* deal with client requests */
int serv_clients(int server_socket)
{
	int clientfd=0,i,found,highest_fd;
	struct timeval timeout;	
	struct clientinfo *fd_list[MAX];
	struct clientinfo *temp_start=NULL;
	pthread_t resolver_thread;
	fd_set connections;

	/* Initialise resolver */
	pthread_create(&resolver_thread,NULL,resolve_thread_create,NULL);

	/* Zero fd set & add server */
	FD_ZERO(&connections);
	FD_SET(server_socket,&connections);
	
	/* Set timeout */
	timeout.tv_sec = SELECT_SECS;
	timeout.tv_usec = 0;
	
	/* Only one fd atm */
	highest_fd = server_socket;

	/* Pad out array of client's fd's */
	for(i=0; i < MAX ;i++)
		fd_list[i]=NULL;
	
	/* Initialise Channel hash table */
	prepare_hash;

	while(signal_caught==0) {
		/* The masterful select() call */
		select(highest_fd+1,&connections,NULL,NULL,&timeout);
		
		/* Check to see if it's a new connection */
		if(FD_ISSET(server_socket,&connections)) {
			new_connection(temp_start,fd_list,&highest_fd,server_socket);
		} else {
			found=0;
			for(i=0;found==0 && i<MAX;i++) {
				if(FD_ISSET(i,&connections)) {
					request(i,fd_list[i],temp_start,&highest_fd,fd_list);
					found=1;
				}
			}
		}
		
		/* Check pings */
		check_pings(fd_list,&highest_fd,temp_start);
			
		/* Reset timer - see select() manpage */
		timeout.tv_sec = SELECT_SECS;
		timeout.tv_usec = 0;
	
		/* Reset fdset */	
		FD_ZERO(&connections);
		FD_SET(server_socket,&connections);
		for(i=0;i<MAX;i++) {
			if(fd_list[i]!=NULL) 
				FD_SET(i,&connections);
		}
		if(highest_fd==-1) 
			highest_fd=server_socket;


	}

	/* Drop all clients, clean up */
	for(i=0;i<MAX;i++) {
		if(fd_list[i]!=NULL)
			shutdown(i,2);
	}
	shutdown(server_socket,2);
}

/* deal with a new connection */
static int new_connection(struct clientinfo *temp_start,
			  struct clientinfo *fd_list[],
			  int *highest_fd,
			  int server_sock)
{
	struct clientinfo *user_info;
	struct sockaddr_in client_info;
	struct hostent *host_info;
	struct clientinfo *current;
	int sin_size,end;
	char buff[20];
	
	sin_size=sizeof(struct sockaddr);
	
	/* Allocate memory for clientinfo struct */
	user_info=malloc(sizeof(struct clientinfo));
	if(user_info==NULL)
		return EXIT_FAILURE;

	/* Accept connection */
	user_info->fd=accept(server_sock,(struct sockaddr *)&client_info,&sin_size);
	if(user_info->fd==-1) {
		printf("Failed to accept\n");
		perror("accept");
		free(user_info);
		return EXIT_FAILURE;
	}

	/* Set non-blocking */
	fcntl(user_info->fd,F_SETFL,O_NONBLOCK);

	/* Update highest fd */
	if(user_info->fd>(*highest_fd)) 
		(*highest_fd)=user_info->fd;	
	
	/* Do a reverse dns to lookup */
	/* must be a better way */
	strncpy(buff,(const char *)inet_ntoa(client_info.sin_addr),20);
	/*	host_info=gethostbyaddr(buff,strlen(buff),AF_INET); */
	host_info =NULL;
	
	if(host_info==NULL) {
		printf("Unable to resolv %s\n",buff);
		perror("gethostbyaddr");
		strncpy(user_info->host,buff,20);
	} else {
		printf("hostname resolved to %s\n",host_info->h_name);
		strncpy(user_info->host,host_info->h_name,HOST_LENGTH);
	}
	
	/* Add the user to the temp list */
	if(temp_start==NULL) {
		temp_start=user_info;
	} else {
		current=temp_start;
		while (end==0) {
			if(current->list_next!=NULL) {
				current=current->list_next;
			} else {
				current->list_next=user_info;
				user_info->list_prev=current;
				end=1;
			}
		}
	}
       	/* Setup default values */	
	user_info->list_next=NULL;	
	fd_list[user_info->fd]=user_info;
	user_info->flags=USR_LIST;
	user_info->lastrecv=time(NULL);
	user_info->nick[0]='\0';
	clients++;
}

/* deals with a client request and calls appropriate command handler */
static int request(int clientfd,
		   struct clientinfo *user_info,
		   struct clientinfo *start,
		   int *highest_fd,
		   struct clientinfo *fd_list[])
{
	int numbytes,i,end,j;
	struct parsed_cmd *p_cmd;
	char buffer[1024];
	
	
	/* Split recieve buffer into individual commands */
	while((numbytes=recv(clientfd,buffer,1024,MSG_PEEK))>0) {
		buffer[numbytes]='\0';
		
		if(numbytes > 1024) {
			/* Flood */
			force_quit(user_info,fd_list,highest_fd,"Excess Flood",start);
			return EXIT_FAILURE;
		} else {
			/* Calculate data for first comamnd */
			for(i=0;buffer[i]!='\n';i++);
		}
		/* Actually Collect data from buffer */
		numbytes=recv(clientfd,buffer,i+1,0);
		buffer[numbytes]='\0';
		/* Check for valid message length */
		if(numbytes > 512) {
			/* Invalid message length */
			/* We may be under attack :( */
			force_quit(user_info,fd_list,highest_fd,"Excess Flood",start);
			return EXIT_FAILURE;
		}
		
		/* Set time we last recieved a packet */
		user_info->lastrecv=time(NULL);	
		
		/* Actually parse and respond to commands */			
		p_cmd=parse_command(buffer);
		
		/* Choose a command to respond with */
		if(p_cmd!=NULL && p_cmd->command!=NULL) {	
			/* Commands that can be excuted when not registered */
			if(lc_strncmp(p_cmd->command,"nick",-1)==0) {
				if(cmd_nick(&user_info,p_cmd,start,fd_list)==EXIT_SUCCESS) {
					user_info->flags |= USR_NICK;
					return move_to_tree(user_info,start);
				} else {
					return EXIT_FAILURE;
				}
			}
			if(lc_strncmp(p_cmd->command,"user",-1)==0) {
				if(cmd_user(user_info,p_cmd)==EXIT_SUCCESS) {
					return move_to_tree(user_info,start);
				} else {
					return EXIT_FAILURE;
				}
			}
			if(lc_strncmp(p_cmd->command,"quit",-1)==0)  {
				return cmd_quit(user_info,p_cmd,fd_list,highest_fd,start);
			}
			/* Commands that can only be executed once registered */
			/* Called via switch_reg_cmd */
			if( !(user_info->flags & USR_LIST) ) {
				return switch_reg_cmd(user_info,p_cmd,start,
						      fd_list,highest_fd);
			} else {
				/* Tell them they need to registered */
				sprintf(buffer,"%s :Register First",p_cmd->command);
				if(user_info->nick[0]!='\0') {
					send_numeric(user_info->fd,user_info->nick,451,buffer);
				} else {
					send_numeric(user_info->fd,"*",451,buffer);
				}
				cmd_free(p_cmd);
				return EXIT_SUCCESS;
			}
				
						      
		}
	}
	/* Check to see if we need to kill the socket */
	if(numbytes <= 0 ) {
		/* Kill socket */
		force_quit(user_info,fd_list,highest_fd,"Connection reset by peer",start);
		return EXIT_FAILURE;
	}
}


/* Switch commands only legal after registration */
/* Called from request() */
static int switch_reg_cmd(struct clientinfo *user_info,
			  struct parsed_cmd *p_cmd,
			  struct clientinfo *start,
			  struct clientinfo *fd_list[],
			  int *highest_fd)
{	  
			
	/* Implement a checking routine for these */
	/* So only registered users can use them */
	if(lc_strncmp(p_cmd->command,"name",-1)==0) {
		return cmd_name(user_info,p_cmd);
	}
	if(lc_strncmp(p_cmd->command,"privmsg",-1)==0) {
		return cmd_privmsg(user_info,p_cmd);
	}
	
	if(lc_strncmp(p_cmd->command,"mode",-1)==0) {
		return cmd_mode(user_info,p_cmd);
	}
	if(lc_strncmp(p_cmd->command,"ping",-1)==0) {
		return cmd_ping(user_info,p_cmd);
	}
	if(lc_strncmp(p_cmd->command,"join",-1)==0) {
		return cmd_join(user_info,p_cmd);
	}
	if(lc_strncmp(p_cmd->command,"pong",-1)==0) {
		/* Remove pingsent flag */
		user_info->flags ^= USR_PINGSENT;
		cmd_free(p_cmd);
		return EXIT_SUCCESS;
	}
	if(lc_strncmp(p_cmd->command,"who",-1)==0) {
		return cmd_who(user_info,p_cmd);
	}
	if(lc_strncmp(p_cmd->command,"part",-1)==0) {
		return cmd_part(user_info,p_cmd);
	}
	if(lc_strncmp(p_cmd->command,"notice",-1)==0) {
		return cmd_notice(user_info,p_cmd);
	}
	if(lc_strncmp(p_cmd->command,"topic",-1)==0) {
		return cmd_topic(user_info,p_cmd);
	}
	if(lc_strncmp(p_cmd->command,"oper",-1)==0) {
		return cmd_oper(user_info,p_cmd);
	}
	if(lc_strncmp(p_cmd->command,"whois",-1)==0) {
		return cmd_whois(user_info,p_cmd);
	}
	 

}

/* move a user into the tree if it's required */
static int move_to_tree(struct clientinfo *user_info,
			struct clientinfo *temp_start)
{
	if(( user_info->flags & (USR_LOGGEDIN ^ USR_LIST ^ USR_NICK)) == (USR_LOGGEDIN ^ USR_LIST ^ USR_NICK)) {
		if(user_info->list_prev==NULL) {
			temp_start=user_info->list_next;
		}
		if(user_info->list_next==NULL) {
			user_info->list_prev->list_next==NULL;
		} 
		if(user_info->list_next!=NULL && user_info->list_prev!=NULL) {
			user_info->list_prev->list_next=user_info->list_next;
			user_info->list_next->list_prev=user_info->list_prev;
		}
		user_info->flags ^= USR_LIST;
		user_info->chan_start=NULL;
		add_client_struct(user_info);
		complete_login(user_info);
		return CLIENT_OK;
	}
	return CLIENT_OK;
}
