#include "network.h"
#include "parse.h"

/* called to setup tcp socket, returns socket or NET_FAILED */
int setup_tcp(struct sockaddr_in * server)
{
	int tcp_socket;
	
	/* Get TCP Socket */
	if((tcp_socket=socket(AF_INET,SOCK_STREAM,0))==-1) {
		return NET_FAILED;
	}
	perror("socket");
	
	/* Bind to socket */
	if(bind(tcp_socket,(struct sockaddr *)server,sizeof(struct sockaddr))==-1) {
		close(tcp_socket);
		return NET_FAILED;
	}
	perror("bind");
	
	/* Listen on socket */
	if(listen(tcp_socket,NET_MAXLISTEN)==-1) {
		close(tcp_socket);
		return NET_FAILED;
	}
	perror("listen");

	/* Set non-blocking */
	fcntl(tcp_socket,F_SETFL,O_NONBLOCK);
	
	return tcp_socket;

}

/* recieves a command and then parses it */
struct parsed_cmd * p_recv(int socket)
{
	char buffer[510];
	int numbytes;
	
	numbytes=recv(socket,buffer,510,0);
	buffer[numbytes]='\0';

	return parse_command(buffer);	
}		

/* sends a msg to a single client */
int send_to_fd(int fd,struct clientinfo *sender,char *command)
{
	char buffer[500];
	int numbytes,buff_len;
	
	/* Form outgoing string */
	ircd_sprintf(buffer,500,":%s!%s@%s %s\n",sender->nick,
						sender->user,
						sender->host,
						command);
			
	buff_len = strlen(buffer);
	numbytes=send(fd,buffer,buff_len,0);
	if(numbytes!=buff_len) {
		return EXIT_FAILURE;
	} else {
		return EXIT_SUCCESS;	
	}
}

int send_raw(int fd,char *buffer)
{
	return send(fd,buffer,strlen(buffer),0);
}


/* sends a numeric reply */
int send_numeric(int fd,char *nick,int code,char *text)
{
	char buffer[550];
	
	ircd_sprintf(buffer,550,":%s %i %s %s\n",server_name,code,nick,text);

	send(fd,buffer,strlen(buffer),0);
}

/* Loop though a channel and send a message */
int send_to_channel(struct clientinfo *sender,
		    struct channel *chan,
		    char *message)
{
	struct user_wrap *curr_client;

	curr_client=chan->users_start;
	while(curr_client!=NULL) {
		send_to_fd(curr_client->user_info->fd,sender,message);
		curr_client=curr_client->next;
	}
}
	
