#include "utility.h"
/* Replacements for strcmp and an interface to calloc */

/* allocates a string that can be free'd when finished with */
char * string_alloc(int length)
{

	return calloc(sizeof(char),length+1);

}

/* replacement for strcmp */
/* lowercase's it before comparing */
/* treats { as lc [,} as lc ],| as lc \ and ^ as lc ~ */
int lc_strncmp(const char *string1,const char *string2,int limit)
{
	char *buffer1;
	char *buffer2;
	int i;
	int retval;
	
	buffer1=string_alloc(strlen(string1));
	buffer2=string_alloc(strlen(string2));

	if(buffer1==NULL || buffer2==NULL) {
		free(buffer1);
		free(buffer2);
		return EXIT_FAILURE;
	}
	
	/* Lowercase strings */
	for(i=0;string1[i]!='\0';i++) {
		if(string1[i] > 64 && string1[i] < 95) {
			buffer1[i]=string1[i] + 32;
		} else {
			buffer1[i]=string1[i];
		}
	}
	buffer1[i]='\0';
		
	for(i=0;string2[i]!='\0';i++) {
		if(string2[i] > 64 && string2[i] < 95) {
			buffer2[i]=string2[i] + 32;
		} else {
			buffer2[i]=string2[i];
		}
	}
	buffer2[i]='\0';

	/* Call strcmp */
	if(limit!=-1) {
		retval = strcmp(buffer1,buffer2);
	} else {
		retval = strncmp(buffer1,buffer2,limit);
	}

	free(buffer1);
	free(buffer2);

	return retval;
}

/* Overflow proof string addition routine */
/* Ha! */
int ircd_strncat(char *dest,const char *source,int length) 
{
	int source_length;
	int dest_length;	
	
	source_length=strlen(source);
	dest_length=strlen(dest);

	if( (length-dest_length) < source_length ) {
		/* Only copy what we can */
		strncat(dest,source,length-dest_length);
		return 0;
	} else {
		strncat(dest,source,source_length);
		return 1;
	}
	
}

/* Overflow proof sprintf for ircd */
/* Oh yeah !!! */
int ircd_sprintf(char *buffer,int buf_size,char *format, ...)
{
	va_list items;
	/* Lengths */
	int mode_len,str_len,max_pow;
	/* Counters */
	int i,j,k;
	unsigned long long int div_by;
	/* Flag */
	int over_flow=0;
	/* Buffers */
	int curr_int;
	long int curr_long;
	char *curr_str;
	/* a quirk of va_arg */
	int curr_char;
	
	mode_len=strlen(format);
	va_start(items,format);

	j=0;	
	for(i=0;i<mode_len && j<buf_size;i++) {
		switch(format[i]) {
			/* if we have to expand into buffer */
			case '%':
				switch(format[i+1]) {
					case 's':
						/* s - String */
						curr_str=va_arg(items,char *);
						str_len=strlen(curr_str);
						if(str_len>(buf_size-j)) {
							over_flow=1;
							i++;
							break;
						}
						for(k=0;k<str_len;k++) {
							buffer[j++]=curr_str[k];
						}
						i++;
						break;
					case 'l':
						/* ld - Long Number */
						curr_long=0;
						curr_long=va_arg(items,long int);
						if(format[i+2]=='d' && curr_long >= LONG_MIN) {
							max_pow=0;
							for(k=0;max_pow==0;k++) {
								div_by=ten_pow(k);
								if((curr_long/div_by)<1) {
									max_pow=k;
								}
							}
							div_by=ten_pow(max_pow)/10;
							for(;div_by>=1;div_by=div_by/10) {
								buffer[j++]=(int) (curr_long % (div_by*10))/div_by +48;
							}
						}
						/* Skip next two */
						i+=2;
						break;

					case 'd':
						/* d - number */
						break;
					case 'c':
						/* c - character */
						curr_char=va_arg(items,int);
						buffer[j++]=curr_char;
						break;
					case 'i':
						/* i - IRCD Error/Reply Code */
						curr_int=va_arg(items,int);
						if(curr_int > 99) {
							buffer[j++]=(int) (curr_int/100) + 48;
							buffer[j++]=(int) ((curr_int%100)/10) + 48;
							buffer[j++]=(int) (curr_int%10) + 48;
							i++;
							break;
						}
						if(curr_int > 9) {
							/* Two digits */
							buffer[j++] = 48;
							/* Be safe :) */
							buffer[j++] =(int) ((curr_int%100)/10)+48;
							buffer[j++] =(int) (curr_int%10) +48;
							i++;
							break;
						} else {
							/* One digit */
							buffer[j++] = 48;
							buffer[j++] = 48;
							buffer[j++] =(int) (curr_int%10)+48;
							i++;
							break;
						}
						
						break;
					default:
						printf("unknown type\n");
					break;
				}
				break;
			case ' ':
				buffer[j++]=' ';
				break;
			
			case '\n':
				buffer[j++]='\n';
				break;

			default:
				buffer[j++]=format[i];
				break;
				
		}
	}
	/* Append NULL character */
#ifdef DEBUG
	printf("j is %d and buf_size is %d\n",j,buf_size);
	printf("over_flow is %d\n",over_flow);
#endif
	buffer[j]='\0';
	va_end(items);

	return over_flow;
}


static unsigned long long int ten_pow(int pow)
{
	int i;
	unsigned long long int result;
	
	if(pow==1) {
		return 1;
	} else {
		result=1;
		for(i=2;i<=pow;i++) {
			result=result * 10;
		}
		return result;
	}
}
